"""
    FreeIOT Core Package Initial Script

    Author: Noah Gao
    Updated at: 2018-02-23
"""
from flask import Flask
from flask_pymongo import PyMongo
from libfreeiot.config import CONFIG

mongo = None

def create_app(config_name, scope = None):
    """
        Function for create Flask App instance
    """
    global mongo
    if scope is None:
        scope = dict()
    app = Flask(__name__) # Initialize app

    # Import project config
    app.config.from_object(CONFIG[config_name])
    CONFIG[config_name].init_app(app)

    app.config["MONGO_URI"] = "mongodb://" + app.config["MONGO_HOST"] + ":" + str(app.config["MONGO_PORT"]) + "/" + app.config["MONGO_DBNAME"]
    print(app.config["MONGO_URI"])

    # Init MongoDB Ext
    mongo = PyMongo(app)

    from libfreeiot.core.routes import create_routes
    app, api = create_routes(app, scope) # Initialize api services with Routes definition
    return app



"""
    The RESTFul resource of device

    Author: Noah Gao
    Updated at: 2018-02-23
"""
from flask import Response, abort
from flask_restful import Resource, reqparse
from flask_jwt_simple import jwt_required
from bson import json_util, ObjectId
from libfreeiot.core import mongo
from libfreeiot.app import scope

class Device(Resource):
    """
        The RESTFul resource class of device
    """
    @jwt_required
    def get(self, device_id=None):
        """
            RESTFul GET Method
        """
        if device_id!=None:
            res = mongo.db.devices.find_one_or_404({'_id': ObjectId(device_id)})
        else:
            res = mongo.db.devices.find()
        return Response(
            json_util.dumps(res),
            mimetype='application/json'
        )

    @jwt_required
    def post(self, device_id=None):
        """
            RESTFul POST Method
        """
        parser = reqparse.RequestParser()
        parser.add_argument('remark', type=str, help='Remark of the device')
        parser.add_argument('status', type=int, help='Status of the device')
        parser.add_argument('version', type=str, help='Version of the device')
        args = parser.parse_args()
        if device_id is None:
            data = args
            res = mongo.db.devices.insert_one(data)
        else:
            data = {}
            if not args["remark"] is None:
                data["remark"] = args["remark"]
            if not args["status"] is None:
                data["status"] = args["status"]
            if not args["version"] is None:
                data["version"] = args["version"]
            if not ObjectId.is_valid(device_id):
                return abort(400)
            if(len(data) < 1):
                return abort(400)
            mongo.db.devices.update_one({"_id": ObjectId(device_id)},{
                "$set": data
            })
        return Response(
            json_util.dumps(data),
            mimetype='application/json'
        )

    @jwt_required
    def delete(self, device_id):
        """
            RESTFul DELETE Method
        """
        res = mongo.db.devices.delete_one({'_id': ObjectId(device_id)})
        return Response(
            json_util.dumps(res.raw_result),
            mimetype='application/json'
        )



"""
    Base Adapter Module
"""
from abc import ABCMeta, abstractmethod

class BaseAdapter(metaclass=ABCMeta):
    """
        Base Adapter Class
    """

    app = None
    scope = dict()

    def init(self, app, scope):
        """
            Initialize Function for Adapter
        """
        self.app = app
        self.scope = scope

    @abstractmethod
    def run(self):
        """
            Main Entry for Adapter
        """
        pass



"""
Routes Module
Author: Noah Gao
Updated at: 2018-2-2
"""
import os
import datetime
from bson import json_util
from flask import request, jsonify, Response
from flask_restful import Api
from flask_jwt_simple import JWTManager, create_jwt
from .resources.device import Device
from .resources.data import Data

JWT_EXPIRES = 7 * 24 * 3600

def create_routes(app, scope = None):
    '''
      Function for create routes
    '''
    if scope is None:
        scope = dict()
    app.config['JWT_SECRET_KEY'] = 'super-secret'  # Change this!
    app.config['JWT_EXPIRES'] = datetime.timedelta(7)
    app.config['UPLOAD_FOLDER'] = os.path.join(os.getcwd(), '/images')
    app.config['MAX_CONTENT_LENGTH'] = 16 * 1024 * 1024
    jwt = JWTManager(app)

    @app.route('/hello')
    def say_hello():
        '''
          Index Route
        '''
        return jsonify({"msg": "Hello World!"})

    @app.route('/api/auth', methods=['POST'])
    def auth():
        '''
          JWT Auth Route
        '''
        username = request.json.get('username', None)
        password = request.json.get('password', None)
        if not username:
            return jsonify({"msg": "Missing username parameter"}), 400
        if not password:
            return jsonify({"msg": "Missing password parameter"}), 400
        if 'auth'in scope.keys():
            authr = scope["auth"].login(username, password)
            if not authr:
                return jsonify({"msg": "Bad username or password"}), 401
        else:
            if username != 'admin' or password == 'admin':
                return jsonify({"msg": "Bad username or password"}), 401
        res = authr or {}
        res["jwt"] = create_jwt(identity=username)
        return Response(
            json_util.dumps(res),
            mimetype='application/json'
        ), 200

    # RESTFul API Routes definition
    api = Api(app)
    api.add_resource(Device, '/api/device', '/api/device/<string:device_id>')
    api.add_resource(Data, '/api/data', '/api/data/<string:data_id>')

    return (app, api)



"""
    FreeIOT Parse Package Initial Script

    Author: Noah Gao
    Updated at: 2018-02-23
"""
import os
from pymongo import MongoClient

mongo = MongoClient(host=os.environ.get("MONGO_HOST") or "localhost", port=int(os.environ.get("MONGO_PORT")) or 27017)
mongo.db = mongo[os.environ.get("MONGO_DBNAME") or "test"]



"""
    The RESTFul resource of device

    Author: Noah Gao
    Updated at: 2018-02-23
"""
import json
from flask import Response
from flask_restful import Resource, reqparse
from flask_jwt_simple import jwt_required
from bson import json_util, ObjectId, DBRef
from libfreeiot.core import mongo

class Data(Resource):
    """
        The RESTFul resource class of data
    """

    def parse_args(self):
        """ Message Parse Method """
        parser = reqparse.RequestParser()
        parser.add_argument('device', type=str, help='ObjectId(str) of the message\'s original device')
        parser.add_argument('topic', type=str, help='The message\'s topic')
        parser.add_argument('flag', type=str, help='The message\'s flag')
        parser.add_argument('content', type=json.loads, help='The message\'s content of JSON Encode')
        args = parser.parse_args()
        args["device"] = DBRef("devices", args["device"])
        return args

    @jwt_required
    def get(self, data_id=None):
        """
            RESTFul GET Method
        """
        if data_id!=None:
            res = mongo.db.datas.find_one_or_404({'_id': ObjectId(data_id)})
        else:
            res = mongo.db.datas.find()
        return Response(
            json_util.dumps(res),
            mimetype='application/json'
        )

    @jwt_required
    def post(self):
        """
            RESTFul POST Method
        """
        args = self.parse_args()
        res = mongo.db.datas.insert_one(args)
        return Response(
            json_util.dumps(res.inserted_id),
            mimetype='application/json'
        )

    @jwt_required
    def put(self, data_id):
        """
            RESTFul PUT Method
        """
        args = self.parse_args()
        res = mongo.db.datas.update_one({"_id": ObjectId(data_id)}, {"$set": args})
        return Response(
            json_util.dumps(res.raw_result),
            mimetype='application/json'
        )

    @jwt_required
    def delete(self, data_id):
        """
            RESTFul DELETE Method
        """
        res = mongo.db.datas.delete_one({'_id': ObjectId(data_id)})
        return Response(
            json_util.dumps(res.raw_result),
            mimetype='application/json'
        )



"""
Message Parse Module
Author: Noah Gao
Updated at: 2018-2-2
"""
from datetime import datetime
from bson import DBRef, ObjectId
from .. import Constant
from . import sys, mongo

def before_check(d_id):
    """ Before Check Function """
    if not ObjectId.is_valid(d_id):
        print(d_id + " is invalid!")
        return
    else:
        d_id = ObjectId(d_id)
    result = mongo.db.devices.find_one({"_id": d_id})
    if result is not None:
        if "status" not in dict(result).keys():
            return False
        if result.get("status") == Constant.Status.STATUS_UNKNOWN:
            r = mongo.db.devices.update_one({"_id": d_id},{ "$set": { "status":  Constant.Status.STATUS_WAIT }})
    return result

def main(client, topic, payload):
    """ Parse Engine Main Function """
    print("Parsing ", topic, payload)
    if topic[0]=='SYS':
        if not before_check(payload["id"]):
            return
        if topic[1]=='online':
            sys.online(client, payload)
        elif topic[1]=='will':
            sys.offline(client, payload)
    else:
        if not before_check(topic[0]):
            return
        res = mongo.db.datas.insert_one({
            "device": DBRef("devices", topic[0]),
            "topic": topic[1],
            "flag": topic[2],
            "content": payload,
            "created_at": datetime.utcnow()
        })
        mongo.db.devices.update_one({"_id": ObjectId(topic[0])},
        {
            "$set": {
                "lastdata." + topic[1]: {
                    "flag": topic[2],
                    "content": payload,
                    "original": DBRef("datas", res.inserted_id),
                    "created_at": datetime.utcnow()
                }
            }
        })
        print(topic, payload, "Created at:", datetime.now())



"""
    File Adapter Module
"""
import time
import uuid
from flask import request, jsonify
from ..base import BaseAdapter

class FileAdapter(BaseAdapter):
    """ File Adapter Class """
    def create_routes(self, app, mongo):
        """ Routes Creator """
        @app.route('/api/upload', methods=['POST'])
        def save_upload():
            """ File Upload POST Route """
            filename = uuid.uuid3(uuid.NAMESPACE_DNS, request.files['file'].filename.split('.')[0] + str(time.time())).hex + "." + request.files['file'].filename.split('.')[1]
            mongo.save_file(filename, request.files['file'], base = "datas")
            return jsonify(filename)

        @app.route('/api/upload/<path:filename>')
        def get_upload(filename):
            """ File Upload GET Route """
            response = mongo.send_file(filename, base = "datas")
            response.cache_control.public = False
            response.cache_control.max_age = 0
            response.cache_control.no_cache = True
            return response

    def run(self):
        """ Main Entry for Adapter """
        self.create_routes(self.app, self.scope["mongo"])
        print('Hello from FileAdapter')



"""
    MQTT Adapter Module
"""
import os
import json
import paho.mqtt.client as mqtt
from ..base import BaseAdapter
from .Parse import main as Parse

class MQTTAdapter(BaseAdapter):
    """ Base Adapter Class """
    client_id = os.environ.get("MQTT_CLIENTID") or "mqtt-adapter"
    server_address = os.environ.get("MQTT_HOST") or "localhost"
    server_port = int(os.environ.get("MQTT_PORT")) or 1883
    client = mqtt.Client(client_id)
    parse_driver = os.environ.get("MQTT_PARSE_DRIVER") or "json"

    def run(self):
        """ Main Entry for Adapter """
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        self.client.connect(self.server_address, self.server_port, 60) # 连接服务器（TCP）
        self.scope["mqttClient"] = self.client # 将 client 代入 Adapter 作用域
        self.client.loop_start()
        print('Hello from MQTTAdapter')

    def on_connect(self, client, userdata, flags, rc):
        """ Callback while conntected """
        print("MQTT Broker connected with result code "+str(rc))
        # Subscribing in on_connect() means that if we lose the connection and
        # reconnect then subscriptions will be renewed.
        self.client.subscribe("SYS/online")
        self.client.subscribe("SYS/will")

    # The callback for when a PUBLISH message is received from the server.
    def on_message(self, client, userdata, msg):
        """ Callback while received messageA """
        Parse.main(client, msg.topic.split('/'), self.parse_driver_select(msg.payload.decode()))

    def parse_driver_select(self, data):
        """ Select a driver to parse data """
        if self.parse_driver == 'msgpack':
            raise OSError("Parse driver 'msgpack' under development.")
        elif self.parse_driver == 'json':
            try:
                return json.loads(data)
            except json.JSONDecodeError as e:
                print("Parsing failed, reason: ", e)
        else:
            raise OSError("Parse driver '" + self.parse_driver + "' under development.")



"""
    Main Entry for FreeIOT

    Author: Noah Gao
    Updated at: 2018-02-23
"""

# Load environment
import os
from dotenv import load_dotenv, find_dotenv
from libfreeiot.core import create_app
from libfreeiot import version

load_dotenv(find_dotenv(usecwd=True), override=True)

print('FreeIOT Version: %s' % version.__version__)
print('Web Service: %s:%s' % (os.environ.get('APP_HOST'), os.environ.get('APP_PORT')))
print('MongoDB Service: %s:%s/%s' % (os.environ.get('MONGO_HOST'), os.environ.get('MONGO_PORT'), os.environ.get('MONGO_DBNAME')))

scope = dict()

def create_flask_app():
    """ Function for create flask application instance """
    global scope
    return create_app(os.getenv('FLASK_CONFIG') or 'default', scope)

def run(port = int(os.environ.get("APP_PORT")),
    host = os.environ.get('APP_HOST', '127.0.0.1'),
    adapters = None,
    app = None):
    """ Main Method for running application """
    global scope

    if app is None:
        app = create_flask_app() # Create Flask Application

    # 代入数据库句柄到 Adapter 作用域
    from .core import mongo
    scope["mongo"] = mongo

    if adapters is None:
        adapters = [] # Create adapter group if not provided
    for adapter in adapters:
        adapter.init(app, scope) # Initialize all adapters
        adapter.run() # Run all adapters

    app.run(debug=os.environ.get("APP_DEBUG") == "true", port=port, host=host, threaded=True) # Start API services



"""
    Library's Setuptools Script
"""
import imp
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

PACKAGE_NAME = 'libfreeiot'
VERSION = imp.load_source('version', os.path.join(here, '%s/version.py' % PACKAGE_NAME)).__version__

setup(
    name = "libfreeiot",
    version = VERSION,
    description = 'A free, open-source IoT Framework',
    author = 'Noah Gao',
    author_email = 'noahgaocn@outlook.com',
    url = 'https://github.com/noahziheng/freeiot',
    download_url = 'https://github.com/noahziheng/freeiot/archive/' + VERSION + '.tar.gz',
    packages = find_packages(),
    install_requires=[    # 依赖列表
        'Flask>=0.12.2',
        'Flask-JWT-Simple>=0.0.3',
        'Flask-PyMongo>=0.5.1',
        'Flask-RESTful>=0.3.6',
        'paho-mqtt>=1.3.1',
        'pylint>=1.8.2',
        'python-dotenv>=0.7.1'
    ]
)



"""
    Default Configuration for Flask Config
"""
import os
from dotenv import load_dotenv, find_dotenv

load_dotenv(find_dotenv(usecwd=True), override=True)

class Config:
    """
        Config base class
    """
    SECRET_KEY = os.environ.get('SECRET_KEY') or 'hard to guess string'
    MONGO_HOST = os.environ.get('MONGO_HOST') or 'localhost'
    MONGO_PORT = int(os.environ.get('MONGO_PORT') or 27017)
    MONGO_DBNAME = os.environ.get('MONGO_DBNAME') or 'test'
    DEBUG = os.environ.get('DEBUG') or True

    @staticmethod
    def init_app(app):
        """
            For Flask Standard
        """
        pass

class DevelopmentConfig(Config):
    """
        Config for debug environment
    """
    DEBUG = True

class TestingConfig(Config):
    """
        Config for test environment
    """
    TESTING = True

class ProductionConfig(Config):
    """
        Config for production environment
    """
    PRODUCTION = True

CONFIG = {
    'development': DevelopmentConfig,
    'testing': TestingConfig,
    'production': ProductionConfig,
    'default': DevelopmentConfig
}



"""
Message Parse Module SYS topic
Author: Noah Gao
Updated at: 2018-2-2
"""
import os
import json
from bson import ObjectId
from .. import Constant
from . import mongo

def online(client, device):
    """ Register Method of device online """
    d_id = ObjectId(dict(device).get("id"))
    mongo.db.devices.update_one({"_id": d_id},{
        "$set": {
            "status":  Constant.Status.STATUS_WAIT,
            "version": dict(device).get("version")
        }
    })
    client.publish(dict(device).get("id") + "/status/d", Constant.Status.STATUS_WAIT)
    if "TOPICS_NEED" in os.environ:
        topics_need = json.loads(os.environ.get("TOPICS_NEED"))
        for item in topics_need:
            client.subscribe(dict(device).get("id") + "/" + item + "/u")
            client.subscribe(dict(device).get("id") + "/" + item + "/d")
            print("Auto subscribing", dict(device).get("id") + "/" + item + "/(u|d)")
    print(dict(device).get("id") + " Online")

def offline(client, device):
    """ Register Method of device offline """
    d_id = dict(device).get("id")
    mongo.db.devices.update_one(
        {"_id": ObjectId(d_id)},
        { "$set": { "status":  Constant.Status.STATUS_UNKNOWN }
        })
    client.publish(dict(device).get("id") + "/status/d", Constant.Status.STATUS_UNKNOWN)
    if "TOPICS_NEED" in os.environ:
        topics_need = json.loads(os.environ.get("TOPICS_NEED"))
        print(topics_need)
        for item in topics_need:
            client.unsubscribe(dict(device).get("id") + "/" + item + "/u")
            client.unsubscribe(dict(device).get("id") + "/" + item + "/d")
    print(dict(device).get("id") + " Offline")
